import pytest

from mistpy.decorators import *
from mistpy.executable_entry import is_mist_function


def test_int_arg():
    params = {'z': 10, 'y': 20}
    x = arg('z', type_hint=int)._decorate([], params)
    assert x == ([], params)

def test_int_arg_with_default():
    params = {}
    x = arg('z', type_hint=int, default = 42)._decorate([], params)
    assert x == ([], {'z': 42})

def test_hello_fn():
    @with_args(
       arg('name', type_hint=str)     
    )
    def hello_fn(name):
        return "Hello " + name

    assert hello_fn(name = 'Allie') == "Hello Allie"
    
def test_hello_fn2():
    @with_args(
       arg('greeting', type_hint=str),     
       arg('name', type_hint=str)     
    )
    def hello_fn2(greeting, name):
        return greeting + " " + name

    assert hello_fn2(name = 'Allie', greeting = 'Hello') == "Hello Allie"

def test_sys_args():
    @on_spark_context
    def sc1(sc):
        return sc

    @on_spark_session
    def sc2(sc):
        return sc

    @on_sql_context
    def sc3(sc):
        return sc

    @on_hive_context
    def sc4(sc):
        return sc

    @on_hive_session
    def sc5(sc):
        return sc

    @on_streaming_context
    def sc6(sc):
        return sc

    fns = [sc1, sc2, sc3, sc4, sc5, sc6]
    for f in fns:
        assert is_mist_function(f) == True
        assert f(42) == 42

def test_complex():
    @with_args(
       arg('a', type_hint=int),     
       arg('b', type_hint=str),
       arg('c', type_hint=list, default = [1,2,3])
    )
    @on_spark_context
    def complex(sc, a, b, c):
        return [sc, a, b, c]

    assert is_mist_function(complex) == True
    assert complex(42, a = 1, b = 'yoyo') == [42, 1, 'yoyo', [1, 2, 3]]



from abc import abstractmethod

from mistpy.tags import TYPE_CHOICES, SPARK_CONTEXT, SPARK_SESSION, SQL_CONTEXT, HIVE_CONTEXT, HIVE_SESSION, SPARK_STREAMING
from mistpy.tags import tags as dec_tags


class BadParameterException(Exception):
    def __init(self, name, type_hint, additional_info=""):
        self.name = name
        self.type_hint = type_hint
        self.additional_info = additional_info

    def __str__(self):
        return "Bad parameter {} of type {}: {}".format(self.name, self.type_hint, self.additional_info)


class SystemArg(object):

    def __init__(self, type_choice, tags=set()):
        if type_choice in TYPE_CHOICES:
            self.type_choice = type_choice
        else:
            raise AttributeError("wrong value of type choice %s".format(type_choice))
        self.tags = tags

    def __call__(self, fn):
        def wrapper(*args, **kwargs):
            args_len = len(args)
            if args_len != 1:
                message = "it is possible to pass only one argument as python arg, but passed %s, " \
                          "all user defined argument passed as kwargs".format(args_len)
                raise BadParameterException(message)
            return fn(args[0], **kwargs)

        wrapper.type_choice = self.type_choice
        wrapper.fn_tags = self.tags
        return wrapper


class AbstractArg(object):
    def __init__(self, name, type_hint, is_optional):
        self.name = name
        self.type_hint = type_hint
        self.is_optional = is_optional

    @abstractmethod
    def _decorate(self, fn_args, fn_kwargs):
        return fn_args, fn_kwargs

    def __call__(self, fn):
        def wrapper(*args, **kwargs):
            new_args, new_kwargs = self._decorate(args, kwargs)
            return fn(*new_args, **new_kwargs)

        return wrapper


class NamedArg(AbstractArg):

    def __init__(self, name, type_hint):
        super(NamedArg, self).__init__(name, type_hint, False)

    def _decorate(self, fn_args, fn_kwargs):
        new_kw = dict(fn_kwargs)
        args = list(fn_args)
        if self.name not in fn_kwargs:
            # TODO < 2 ?
            if len(args) < 2:  # one for context
                raise BadParameterException("not found argument in kwargs by %s and by args order".format(self.name))
            else:
                new_kw[self.name] = args[1]
                args.pop(1)
        return args, new_kw


class NamedArgWithDefault(AbstractArg):
    def __init__(self, name, type_hint, default=None):
        super(NamedArgWithDefault, self).__init__(name, type_hint, True)
        self.default = default

    def _decorate(self, fn_args, fn_kwargs):
        if self.name in fn_kwargs:
            return fn_args, fn_kwargs
        else:
            fn_kwargs[self.name] = self.default
            return fn_args, fn_kwargs


class OptionalArg(AbstractArg):
    def __init__(self, name, type_hint):
        super(OptionalArg, self).__init__(name, type_hint, True)

    def _decorate(self, fn_args, fn_kwargs):
        new_kwargs = dict(fn_kwargs)
        if self.name not in fn_kwargs:
            new_kwargs[self.name] = None
        return fn_args, fn_kwargs


def arg(name, type_hint, default=None, optional=False):
    if optional:
        return OptionalArg(name, type_hint)
    if default is not None:
        return NamedArgWithDefault(name, type_hint, default)
    return NamedArg(name, type_hint)


def opt_arg(name, type_hint):
    return arg(name, type_hint, optional=True)


class __complex_type(object):
    def __init__(self, container_type, main_type):
        self.container_type = container_type
        self.main_type = main_type


def list_type(main_type):
    return __complex_type(list, main_type)


def with_args(*args_decorator):
    def _call(f):
        def _wraps(*args, **kwargs):
            for arg_decorator in args_decorator:
                fn = arg_decorator(f)
            return fn(*args, **kwargs)
        
        _wraps.__dict__.update(f.__dict__)
        _wraps.args_def = args_decorator
        return _wraps

    return _call


on_spark_context = SystemArg(SPARK_CONTEXT)
on_spark_session = SystemArg(SPARK_SESSION, dec_tags.sql)
on_sql_context = SystemArg(SQL_CONTEXT, dec_tags.sql)
on_hive_context = SystemArg(HIVE_CONTEXT, dec_tags.hive)
on_hive_session = SystemArg(HIVE_SESSION, dec_tags.hive)
on_streaming_context = SystemArg(SPARK_STREAMING, dec_tags.streaming)



import inspect
import sys
import importlib
import types
from abc import abstractmethod
from collections import namedtuple
from inspect import isfunction

from mistpy.tags import SPARK_CONTEXT, SPARK_SESSION, HIVE_SESSION, HIVE_CONTEXT, SQL_CONTEXT
from mistpy.tags import tags as dec_tags


class ExecutableEntry(object):

    def __init__(self, type_choice, args=None, tags=None):
        self._type_choice = type_choice
        if args is None:
            args = []

        if tags is None:
            tags = set()

        self._args = args
        self._tags = tags

    @property
    def args_info(self):
        return self._args

    @property
    def selected_spark_argument(self):
        return self._type_choice

    @property
    def tags(self):
        return self._tags

    @abstractmethod
    def invoke(self, context_wrapper, params):
        pass


class FunctionEntry(ExecutableEntry):
    def __init__(self, fn):
        self._fn = fn
        if hasattr(fn, 'args_def'):
            args = fn.args_def
        else:
            args = []
        type_choice = fn.type_choice
        _tags = fn.fn_tags
        super(FunctionEntry, self).__init__(type_choice, args, _tags)

    def invoke(self, context_wrapper, params):
        return self._fn(context_wrapper, **params)


ArgInfo = namedtuple('ArgInfo', ['name', 'type_hint'])


def extract_args_from_method_py2(method):
    args_spec = inspect.getargspec(method)
    user_args = args_spec.args[1:]
    return list(map(lambda name: ArgInfo(name, None), user_args))


def extract_args_from_method_py3(method):
    sign = inspect.signature(method)
    result = []
    user_params = list(sign.parameters.items())[1:]
    for k, v in user_params:
        result.append(ArgInfo(k, None))
    return result


def extract_args_from_method(method):
    if sys.version_info[0] == 2:
        args = extract_args_from_method_py2(method)
    else:
        args = extract_args_from_method_py3(method)
    return args


def is_mist_function(fn):
    return hasattr(fn, 'fn_tags') and hasattr(fn, 'type_choice')


def get_metadata(fn):
    if isfunction(fn):
        if is_mist_function(fn):
            return FunctionEntry(fn)
        else:
            raise Exception(fn.__name__ + ' is not a mist function')

def load_egg_entry(path, fn_name):
    sys.path.append(path)
    splitted = fn_name.split('.')
    module_path = '.'.join(splitted[:-1])
    fn_attr = splitted[-1]

    module = importlib.import_module(module_path)
    return getattr(module, fn_attr)

def load_one_file(path, fn_name):
    with open(path) as file:
        code = compile(file.read(), path, "exec")
    user_job_module = types.ModuleType("<user_job>")
    exec(code, user_job_module.__dict__)
    return getattr(user_job_module, fn_name)

def load_entry_object(path, fn_name):
    if (path.endswith(".egg")):
        return load_egg_entry(path, fn_name)
    else:
        return load_one_file(path, fn_name)

def load_entry(path, fn_name):
    module_entry = load_entry_object(path, fn_name)
    return get_metadata(module_entry)


# coding=utf-8 

from pyspark.conf import SparkConf
from pyspark.context import SparkContext

class ContextWrapper(object):

    def __init__(self):
        pass

    def set_context(self, java_gateway):
        spark_context_wrapper = java_gateway.entry_point.sparkContextWrapper()
        j_spark_conf = spark_context_wrapper.sparkConf()
        p_spark_conf = SparkConf(_jvm=java_gateway.jvm, _jconf=j_spark_conf)
        j_spark_context = spark_context_wrapper.javaContext()
        self._context = SparkContext(jsc=j_spark_context, gateway=java_gateway, conf=p_spark_conf)

    def set_sql_context(self, java_gateway):
        from pyspark.sql import SQLContext
        spark_context_wrapper = java_gateway.entry_point.sparkContextWrapper()
        self._sql_context = SQLContext(self._context, sparkSession = spark_context_wrapper.sparkSession(False), jsqlContext = spark_context_wrapper.sqlContext())

    def set_hive_context(self, java_gateway):
        from pyspark.sql import HiveContext
        spark_context_wrapper = java_gateway.entry_point.sparkContextWrapper()
        self._hive_context = HiveContext(self._context, spark_context_wrapper.hiveContext())

    def set_session(self, java_gateway):
        from pyspark.sql import SparkSession
        self._session = SparkSession.builder.config(conf=self._context.getConf()).getOrCreate()

    def set_hive_session(self, java_gateway):
        from pyspark.sql import SparkSession
        self._session = SparkSession.builder.config(conf=self._context.getConf()).enableHiveSupport().getOrCreate()

    def set_streaming_context(self, java_gateway):
        from pyspark.streaming import StreamingContext
        spark_context_wrapper = java_gateway.entry_point.sparkContextWrapper()
        self._streaming_context = StreamingContext(self._context, java_gateway.entry_point.sparkStreamingWrapper().getDurationSeconds())

    @property
    def context(self):
        return self._context

    @property
    def sql_context(self):
        return self._sql_context

    @property
    def hive_context(self):
        return self._hive_context

    @property
    def session(self):
        return self._session

    @property
    def streaming_context(self):
        return self._streaming_context



# coding=utf-8

import sys, traceback, argparse
from py4j.java_gateway import java_import, JavaGateway, GatewayClient
from py4j.java_collections import JavaMap, JavaList

from pyspark.sql.types import *

from mistpy.executable_entry import load_entry
from mistpy.context_wrapper import ContextWrapper
from mistpy.decorators import SPARK_CONTEXT, SPARK_SESSION, SPARK_STREAMING, HIVE_CONTEXT, HIVE_SESSION, SQL_CONTEXT


def to_python_types(any):
    python_any = any
    if isinstance(any, JavaMap):
        python_any = dict()
        for key, value in any.items():
            python_any[key] = to_python_types(value)
    elif isinstance(any, JavaList):
        python_any = list()
        for i, value in enumerate(any):
            python_any.insert(i, to_python_types(value))
    return python_any


def initialized_context_value(_gateway, context_wrapper, selected_spark_argument):
    if selected_spark_argument == SPARK_CONTEXT:
        argument = context_wrapper.context
    elif selected_spark_argument == SPARK_SESSION:
        context_wrapper.set_session(_gateway)
        argument = context_wrapper.session
    elif selected_spark_argument == HIVE_SESSION:
        context_wrapper.set_hive_session(_gateway)
        argument = context_wrapper.session
    elif selected_spark_argument == HIVE_CONTEXT:
        context_wrapper.set_hive_context(_gateway)
        argument = context_wrapper.hive_context
    elif selected_spark_argument == SPARK_STREAMING:
        context_wrapper.set_streaming_context(_gateway)
        argument = context_wrapper.streaming_context
    elif selected_spark_argument == SQL_CONTEXT:
        context_wrapper.set_sql_context(_gateway)
        argument = context_wrapper.sql_context
    else:
        raise Exception('Unknown spark argument type: ' + selected_spark_argument)
    return argument


def execution_cmd(args):
    _client = GatewayClient(port=args.gateway_port)
    _gateway = JavaGateway(_client, auto_convert=True)
    _entry_point = _gateway.entry_point

    java_import(_gateway.jvm, "org.apache.spark.SparkContext")
    java_import(_gateway.jvm, "org.apache.spark.SparkEnv")
    java_import(_gateway.jvm, "org.apache.spark.SparkConf")
    java_import(_gateway.jvm, "org.apache.spark.streaming.*")
    java_import(_gateway.jvm, "org.apache.spark.streaming.api.java.*")
    java_import(_gateway.jvm, "org.apache.spark.streaming.api.python.*")
    java_import(_gateway.jvm, "org.apache.spark.api.java.*")
    java_import(_gateway.jvm, "org.apache.spark.api.python.*")
    java_import(_gateway.jvm, "org.apache.spark.mllib.api.python.*")
    java_import(_gateway.jvm, "org.apache.spark.*")
    java_import(_gateway.jvm, "org.apache.spark.sql.*")
    java_import(_gateway.jvm, 'java.util.*')

    context_wrapper = ContextWrapper()
    context_wrapper.set_context(_gateway)

    configuration_wrapper = _entry_point.configurationWrapper()
    error_wrapper = _entry_point.error()
    path = configuration_wrapper.path()
    fn_name = configuration_wrapper.className()
    parameters = configuration_wrapper.parameters()

    data_wrapper = _entry_point.data()


    try:
        executable_entry = load_entry(path, fn_name)
        selected_spark_argument = executable_entry.selected_spark_argument

        argument = initialized_context_value(_gateway, context_wrapper, selected_spark_argument)
        result = executable_entry.invoke(argument, to_python_types(parameters))
        data_wrapper.set(result)

    except Exception:
        error_wrapper.set(traceback.format_exc())



SPARK_CONTEXT = 'sp_context'
SPARK_SESSION = 'sp_session'
SQL_CONTEXT = 'sql_context'
HIVE_CONTEXT = 'hive_context'
HIVE_SESSION = 'hive_session'
SPARK_STREAMING = 'streaming'

TYPE_CHOICES = (SPARK_CONTEXT, SPARK_SESSION, SQL_CONTEXT, HIVE_CONTEXT, HIVE_SESSION, SPARK_STREAMING)


class __tags(object):
    def __init__(self):
        self.sql = {'sql'}
        self.hive = {'hive', 'sql'}
        self.streaming = {'streaming'}


tags = __tags()



# coding=utf-8

import argparse
import traceback

from py4j.java_gateway import java_import, JavaGateway, GatewayClient

from mistpy.decorators import __complex_type
from mistpy.executable_entry import load_entry


def to_scala_arg_type(type_hint, gateway):
    if type_hint is None:
        return gateway.jvm.PythonUtils.anyType()

    if isinstance(type_hint, __complex_type):
        if issubclass(type_hint.container_type, list):
            underlying = to_scala_arg_type(type_hint.main_type, gateway)
            arg_type = gateway.jvm.PythonUtils.listType(underlying)
        else:
            arg_type = gateway.jvm.PythonUtils.anyType()
    else:
        if issubclass(type_hint, str):
            arg_type = gateway.jvm.PythonUtils.strType()
        elif issubclass(type_hint, int):
            arg_type = gateway.jvm.PythonUtils.intType()
        elif issubclass(type_hint, float):
            arg_type = gateway.jvm.PythonUtils.doubleType()
        elif issubclass(type_hint, bool):
            arg_type = gateway.jvm.PythonUtils.boolType()

        elif issubclass(type_hint, list):
            arg_type = gateway.jvm.PythonUtils.listType(gateway.jvm.PythonUtils.anyType())
        else:
            arg_type = gateway.jvm.PythonUtils.anyType()

    return arg_type


def to_scala_internal_arg_info(tags, gateway):
    return gateway.jvm.PythonUtils.systemArg(tags)


def to_scala_arg_info(arg_decorator_like, gateway):
    type_hint = arg_decorator_like.type_hint
    arg_type = to_scala_arg_type(type_hint, gateway)
    if arg_decorator_like.is_optional:
        arg_type = gateway.jvm.PythonUtils.optType(arg_type)
    arg = gateway.jvm.PythonUtils.userArg(arg_decorator_like.name, arg_type)
    return arg


def metadata_cmd(input_args):
    _client = GatewayClient(port=input_args.gateway_port)
    _gateway = JavaGateway(_client, auto_convert=True)
    _entry_point = _gateway.entry_point
    java_import(_gateway.jvm, 'java.util.*')
    java_import(_gateway.jvm, 'io.hydrosphere.mist.python.PythonUtils')
    error_wrapper = _entry_point.error()
    fn_name = _entry_point.functionName()
    file_path = _entry_point.filePath()
    data_wrapper = _entry_point.data()

    try:
        executable_entry = load_entry(file_path, fn_name)
        internal_arg = to_scala_internal_arg_info(executable_entry.tags, _gateway)
        arg_infos = list(map(lambda a: to_scala_arg_info(a, _gateway), executable_entry.args_info))
        arg_infos.append(internal_arg)
        data_wrapper.set(_gateway.jvm.PythonUtils.toSeq(arg_infos))
    except Exception:
        error_wrapper.set(traceback.format_exc())



import pytest

from mistpy.decorators import *
from mistpy.executable_entry import is_mist_function, get_metadata

def test_no_args():
    @on_spark_context
    def sc1(sc):
        return sc

    meta = get_metadata(sc1)
    assert meta.args_info == set() 
    



import os
from setuptools import setup, find_packages

here = os.path.abspath(os.path.dirname(__file__))

with open('README.rst', 'r') as f:
    readme = f.read()


about = {}
with open(os.path.join(here, 'mistpy', '__version__.py'), 'r') as f:
    exec(f.read(), about)

setup(
    name = 'mistpy',
    version= about['__version__'],
    description = 'Mist python library api',
    long_description=readme,
    author= 'Hydrospheredata',
    author_email = 'info@hydrosphere.io',
    url = 'https://hydrosphere.io/mist',
    license = 'Apache 2.0',

    package_dir={'mistpy': 'mistpy'},
    include_package_data=True,
    zip_safe=False,
    classifiers=(
        'Intended Audience :: Developers',
        'Natural Language :: English',
        'License :: OSI Approved :: Apache Software License',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
    ),

    packages=find_packages(exclude=['tests']),
    setup_requires=['pytest-runner'], 
    tests_require=['pytest'],
    test_suite='tests'
)
